/**************************************************************************
 *
 * Source code to extract the Local Optima Network (LON) of the QAP
 *
 * Source   : QAPnetwork.cpp
 * Language : C++
 * Author   : Sebastien Verel
 * Version  : 0.1
 * Date     : 07/30/2010
 *
 * Related publications:
 *
 * Daolio F., Tomassini M., Verel S., Ochoa G.
 * Communities of Minima in Local Optima Networks of Combinatorial Spaces
 * Physica A: Statistical Mechanics and its Applications, Volume 390, Issue 9, Pages 1684-1694, May 2011.
 *
 * Daolio F., Verel S., Ochoa G. & Tomassini M. (2010). 
 * Local Optima Networks of the Quadratic Assignement Problem. 
 * WCCI 2010 IEEE World Congress on Computational Intelligence (pp. 3145-3152). IEEE Press.
 *
 *
 * To compile:
 *   The GNU Scientific Library (GSL) is required. See the web site : http://www.gnu.org/s/gsl/ 
 * 
 *   Then type the classical commands:
 *      g++ -c qapNetwork.cpp -I.
 *      g++ -o qapNetwork qapNetwork.o -L. -lm -lgsl
 *
 * To execute:
 *   ./qapNetwork instances/KCso7uni-1.dat out.nodes out.edges   
 *   or
 *   ./qapNetwork instances/KCso7uni-1.dat out.nodes out.edges out.fitness
 *
 * "instances/KCso7uni-1.dat" is an instance file name of the QAP problem to consider in QAPlib format
 * some instances are given in the directory instances
 *
 * "out.nodes" is the file with the id, - fitness, and the basin size of each node (local optimum) of the network :
 *   idNode1 fitness1 basinSize1
 *   idNode2 fitness2 basinSize2
 *   ....
 * warning: the computed fitness is the opposite fitness (- fitness) to tranform the QAP into a maximization problem
 *
 * "out.edges" is the file with weighted edges of the network:
 *   idNode_i1 idNode_j1 weight_ij1 idNode_j2 wij2 idNode_j3 weight_ij3 ...
 *   ....
 *
 *  which means that there is the edge (idNode_i1, idNode_j1) with the weight weight_ij1, etc.
 *
 *
 * Warnings: 
 *   In this version, the full enumeration of the search space is computed.
 *   The search space have to be 'small'.
 *
 * This code could be used for other permutation search space (TSP, Flow shop, etc.)
 *
 * Contact:
 *   Sebastien Verel, sebastien.verel@unice.fr
 *
 **************************************************************************/

#ifndef __abrBasinH
#define __abrBasinH

#include <basin.h>
#include <bestNeighbor.h>
#include <vector>

/***************************************************************
*
* simple binary tree
* to have all the basins without repetition
*
***************************************************************/

template <class Telem>
class _Node 
{
public:
    Telem * basin;

    _Node<Telem> * leftNode;
    _Node<Telem> * rightNode;


    /***************************************************************
    *
    * add in order without repetition
    * if the node exist, the size of basin is increases and the edges is updated
    *
    * true if the value is added
    ***************************************************************/
    bool add(Telem * b) {
	if (basin->id != b->id) 
	    if (b->id < basin->id) 
		if (leftNode == NULL) {
		    leftNode = new _Node<Telem>;
		    
		    leftNode->basin = b;
		    
		    leftNode->leftNode  = NULL;
		    leftNode->rightNode = NULL;	  
		    
		    return true;
		} else
		    return leftNode->add(b);
	    else
		if (rightNode == NULL) {
		    rightNode = new _Node<Telem>;
		    
		    rightNode->basin = b;
		    
		    rightNode->leftNode  = NULL;
		    rightNode->rightNode = NULL;	  
	  
		    return true;
		} else
		    return rightNode->add(b);
	else { 
	    return false;
	}
    }

  /***************************************************************
   *
   * del all the subtree from this node
   *
   ***************************************************************/
  void del() {
    if (leftNode != NULL) {
      leftNode->del();

      delete leftNode;

      leftNode = NULL;
    }

    if (rightNode != NULL) {
      rightNode->del();
      
      delete rightNode;

      rightNode = NULL;
    }
  }

};

// simple Abre binaire de recherche to insert in increase order

template <class Telem>
class Abr
{
public:

  /***************************************************************
   *
   * empty constructor
   *
   ***************************************************************/
  Abr() 
  {
    _root = NULL;
    _size = 0;
  };

  /***************************************************************
   *
   * destructor
   *
   ***************************************************************/
  ~Abr() 
  {
    del();
  };

  /***************************************************************
   *
   * true if the stack is empty
   *
   ***************************************************************/
  bool empty() {
    return (_root == NULL);
  }
    
  /***************************************************************
   *
   * return the value on the root
   *
   ***************************************************************/
  unsigned root() {
    if (!empty())       
      return _root->basin->id;
    else
      cerr << "Abr.root: error empty tree" << endl;
  }

  /***************************************************************
   *
   * add the value in the increase order in the tree
   * without duplication
   *
   ***************************************************************/
  bool add(Telem * b) {
    if (_root == NULL) {
	_root = new _Node<Telem>;

	_root->basin = b;
      
	_root->leftNode  = NULL;
	_root->rightNode = NULL;	  
      
	_size = 1;

	return true;
    } else {
	if (_root->add(b)) {     
	    _size++;
	    return true;
	} else
	    return false;
    }
  }

  /***************************************************************
   *
   * return all the tree in increase order into vector
   *
   ***************************************************************/
  void get(vector<Telem *> & basins) {
    basins.resize(_size);

    __pos = 0;
    getPrefixe(basins, _root);

  }

  /***************************************************************
   *
   * print the tree
   *
   ***************************************************************/
  void print() {
    if (_root != NULL)
      cout << "root : " << _root->basin->id << endl;
    else
      cout << "root : empty" << endl;
    printPrefixe(_root);
  }

  /***************************************************************
   *
   * print the tree
   *
   ***************************************************************/
  void printPrefixe(_Node<Telem> * node) {
    if (node != NULL) {
      cout << "(" ;
      printPrefixe(node->leftNode);
      
      cout << ", ";
      cout << node->basin->id ;
      cout << ", ";

      printPrefixe(node->rightNode);
      cout << ")";
    }
  }

  /***************************************************************
   *
   * return basin pointer od _id if the _id exist in the abr
   * otherwise return NULL
   *
   ***************************************************************/
  Telem * find(unsigned _id) {
    __id = _id;

    return isIn(_root);
  }

  /***************************************************************
   *
   * size of the stack
   *
   ***************************************************************/
  unsigned size() {
    return _size;
  }

  /***************************************************************
   *
   * delete the tree from "node"
   *
   ***************************************************************/
  void del() {
    if (_root != NULL) {
      _root->del();

      delete _root;
      _root = NULL;

      _size = 0;
    }
  }

private:
  /***************************************************************
   *
   * tree in prefixe order into vector "values"
   * initialize the index of each node
   *
   ***************************************************************/
  void getPrefixe(vector<Basin *> & basins, _Node<Telem> * node) {
    if (node != NULL) {
      getPrefixe(basins, node->leftNode);

//      node->p->position = __pos;
      basins[__pos] = node->basin;
      __pos++;

      getPrefixe(basins, node->rightNode);
     }
  }

  /***************************************************************
   *
   * tree in prefixe order into vector "values"
   * initialize the index of each node
   *
   ***************************************************************/
  void getPrefixe(vector<BestNeighbor *> & basins, _Node<BestNeighbor> * node) {
    if (node != NULL) {
      getPrefixe(basins, node->leftNode);

      node->basin->position = __pos;
      basins[__pos] = node->basin;
      __pos++;

      getPrefixe(basins, node->rightNode);
     }
  }

  /***************************************************************
   *
   * return the basin pointer if "__id" is in the tree
   *
   ***************************************************************/
  Telem * isIn(_Node<Telem> * node) {
    if (node != NULL) {
      if (node->basin->id == __id)
	return node->basin;
      else
	if (node->basin->id < __id)
	  return isIn(node->rightNode);
	else
	  return isIn(node->leftNode);	  
    } else
      return NULL;
  }


  // root of the tree
  _Node<Telem> * _root;

  // size of the tree
  unsigned _size;

  // tempory value
  unsigned __id;
  unsigned __pos;
};

#endif

// Local Variables:
// coding: iso-8859-1
// mode: C++
// c-file-offsets: ((c . 0))
// c-file-style: "Stroustrup"
// fill-column: 80
// End:

