/**************************************************************************
 *
 * Source code to extract the Local Optima Network (LON) of the QAP
 *
 * Source   : QAPnetwork.cpp
 * Language : C++
 * Author   : Sebastien Verel
 * Version  : 0.1
 * Date     : 07/30/2010
 *
 * Related publications:
 *
 * Daolio F., Tomassini M., Verel S., Ochoa G.
 * Communities of Minima in Local Optima Networks of Combinatorial Spaces
 * Physica A: Statistical Mechanics and its Applications, Volume 390, Issue 9, Pages 1684-1694, May 2011.
 *
 * Daolio F., Verel S., Ochoa G. & Tomassini M. (2010). 
 * Local Optima Networks of the Quadratic Assignement Problem. 
 * WCCI 2010 IEEE World Congress on Computational Intelligence (pp. 3145-3152). IEEE Press.
 *
 *
 * To compile:
 *   The GNU Scientific Library (GSL) is required. See the web site : http://www.gnu.org/s/gsl/ 
 * 
 *   Then type the classical commands:
 *      g++ -c qapNetwork.cpp -I.
 *      g++ -o qapNetwork qapNetwork.o -L. -lm -lgsl
 *
 * To execute:
 *   ./qapNetwork instances/KCso7uni-1.dat out.nodes out.edges   
 *   or
 *   ./qapNetwork instances/KCso7uni-1.dat out.nodes out.edges out.fitness
 *
 * "instances/KCso7uni-1.dat" is an instance file name of the QAP problem to consider in QAPlib format
 * some instances are given in the directory instances
 *
 * "out.nodes" is the file with the id, - fitness, and the basin size of each node (local optimum) of the network :
 *   idNode1 fitness1 basinSize1
 *   idNode2 fitness2 basinSize2
 *   ....
 * warning: the computed fitness is the opposite fitness (- fitness) to tranform the QAP into a maximization problem
 *
 * "out.edges" is the file with weighted edges of the network:
 *   idNode_i1 idNode_j1 weight_ij1 idNode_j2 wij2 idNode_j3 weight_ij3 ...
 *   ....
 *
 *  which means that there is the edge (idNode_i1, idNode_j1) with the weight weight_ij1, etc.
 *
 *
 * Warnings: 
 *   In this version, the full enumeration of the search space is computed.
 *   The search space have to be 'small'.
 *
 * This code could be used for other permutation search space (TSP, Flow shop, etc.)
 *
 * Contact:
 *   Sebastien Verel, sebastien.verel@unice.fr
 *
 **************************************************************************/
#ifndef __coupleListH
#define __coupleListH


/***************************************************************
 *
 * simple stack of unsigned int
 * with pointer
 *
 ***************************************************************/

class _CoupleNode 
{
public:
  _CoupleNode() {
    n = 0;
    p = 0.0;

    next = NULL;
  }

  _CoupleNode(unsigned _n, double _p) {
    n = _n;
    p = _p;
    
    next = NULL;
  }

  unsigned n;
  double   p;

  _CoupleNode * next;

};

class CoupleList
{
public:
  
  /***************************************************************
   *
   * empty constructor
   *
   ***************************************************************/
  CoupleList() 
  {
    _head  = NULL;
    _size = 0;
  };

  /***************************************************************
   *
   * destructor
   *
   ***************************************************************/
  ~CoupleList() 
  {
    while (!empty()) 
      pop();
  };

  /***************************************************************
   *
   * true if the list is empty
   *
   ***************************************************************/
  bool empty() {
    return (_head == NULL);
  }
    
  /***************************************************************
   *
   * delete all the list
   *
   ***************************************************************/
  void del() {
    while (!empty()) 
      pop();
  }
    
  /***************************************************************
   *
   * delete the head of the list if it is possible
   *
   ***************************************************************/
  void pop() {
    if (!empty()) {
      _CoupleNode * n = _head->next;
      delete _head;
      _head = n;
      _size--;
    } else
      cerr << "CoupleList.pop: error empty list" << endl;
  }
  
  /***************************************************************
   *
   * return the unsigned value of the head of the list
   *
   ***************************************************************/
  unsigned number() {
    if (!empty())       
      return _head->n;
    else
      cerr << "CoupleList.number: error empty list" << endl;
  }
  
  /***************************************************************
   *
   * return the double value of the head of the list
   *
   ***************************************************************/
  double prob() {
    if (!empty())       
      return _head->p;
    else
      cerr << "CoupleList.prob: error empty list" << endl;
  }
  
  /***************************************************************
   *
   * add a couple of the head
   *
   ***************************************************************/
  void cons(unsigned num, double pr) {
    _CoupleNode * node = new _CoupleNode;
    
    node->n    = num;
    node->p    = pr;
    node->next = _head;
    
    _head = node;
    
    _size++;
  }

  /***************************************************************
   *
   * print the list
   *
   ***************************************************************/
  void print() {
    _CoupleNode * node = _head;
    
    while (node != NULL) {
      cout << " " << node->n << " " << node->p;
      node = node->next;
    }
    
    //    cout << endl;
  }

  /***************************************************************
   *
   * print the list
   *
   ***************************************************************/
  void print(fstream & file) {
    _CoupleNode * node = _head;
    
    while (node != NULL) {
      file << " " << node->n << " " << node->p;
      node = node->next;
    }
    
    //    cout << endl;
  }

  /***************************************************************
   *
   * add a couple of the head
   * with respect of increase number
   *
   ***************************************************************/
  void consOrder(unsigned num, double pr) {
    _CoupleNode * node = new _CoupleNode;

    node->n    = num;
    node->p    = pr;

    if (_head == NULL || _head->n > num) {
      node->next = _head;
      _head = node;

      _size++;
    } else {
      if (_head->n == num) {
	_head->p = pr + _head->p;
	delete node;
      } else {
	_CoupleNode * iter = _head->next;
	_CoupleNode * pred = _head;

	while (iter != NULL && iter->n < num) {
	  iter = iter->next;
	  pred = pred->next;
	}
	
	if (iter != NULL && iter->n == num) {
	  iter->p = pr + iter->p;
	  delete node;
	} else {
	  pred->next = node;
	  node->next = iter;
	  
	  _size++;
	}
      }
    }
  }

  /***************************************************************
   *
   * add the basin of attraction "basinV" (with coeffcient alpha)
   * to the current basin 
   * 
   *  keeping the order of "n" of the list basin
   *
   ***************************************************************/
  void add(double alpha, CoupleList & basinV) {
    // place the iterator at the begining of the lists
    initIterator();
    basinV.initIterator();

    // while there is something to add to basin
    while (!basinV.emptyIterator()) {
      if (emptyIterator()) {       // basin is empty, so add node of basinV to basin 
	consIterator(basinV.iterator()->n, alpha * basinV.iterator()->p);
	basinV.nextIterator();
      } else
	if (basinV.iterator()->n < iterator()->n) { // the value of basinV is lower 
	  consIterator(basinV.iterator()->n, alpha * basinV.iterator()->p);
	  basinV.nextIterator();
	} else
	  if (iterator()->n == basinV.iterator()->n) {
	    iterator()->p += alpha * basinV.iterator()->p;
	    nextIterator();
	    basinV.nextIterator();
	  } else {
	    nextIterator();
	  }
    }

  }

  /***************************************************************
   *
   * size of the list
   *
   ***************************************************************/
  unsigned size() {
    return _size;
  }

  /***************************************************************
   *
   * head pointer of the list
   *
   ***************************************************************/
  _CoupleNode * head() {
    return _head;
  }

  /***************************************************************
   *
   * iterator take place at the head
   *
   ***************************************************************/
  void initIterator() {
    _iterator       = _head;
    _beforeIterator = _head;
  }

  /***************************************************************
   *
   * place the iterator to the next place
   *
   ***************************************************************/
  void nextIterator() {
    if (_iterator != NULL) {
      if (_iterator == _head) {
	_iterator       = _head->next;
	_beforeIterator = _head;
      } else {
	_iterator       = _iterator->next;
	_beforeIterator = _beforeIterator->next;
      }
    } else
      cerr << "coupleList.nextIterator: there is no next pointer after the iterator" << endl;
  }

  /***************************************************************
   *
   * add a node before the iterator if the iterator is not NULL
   * add at the end of list otherwise
   *
   ***************************************************************/
  void consIterator(unsigned n, double p) {
    _CoupleNode * node = new _CoupleNode(n, p);

    if (_iterator != NULL) {
      if (_iterator == _head) {
	node->next = _head;
	_head = node;

	_iterator       = _head->next;
	_beforeIterator = _head;
      } else {
	_beforeIterator->next = node;
	node->next            = _iterator;

	_beforeIterator = _beforeIterator->next;
      }
    } else {
      if (_beforeIterator != NULL) {
	_beforeIterator->next = node;
	node->next = NULL;

	_beforeIterator = _beforeIterator->next;
      } else {
	_head = node;
	node->next = NULL;
	
	_iterator = NULL;
	_beforeIterator = _head;
      }
    }

    _size++;
  }


  /***************************************************************
   *
   * give the iterator pointer
   *
   ***************************************************************/
  _CoupleNode * iterator() {
    return _iterator;
  }

  /***************************************************************
   *
   * return true if the iterator is at the end of the list
   *
   ***************************************************************/
  bool emptyIterator() {
    return (_iterator == NULL);
  }


private:
  // the begin of the list
  _CoupleNode * _head;

  // iterator of the list
  _CoupleNode * _iterator;
  _CoupleNode * _beforeIterator;

  // size of the list
  unsigned _size;

};

#endif
