/**************************************************************************
 *
 * Source code to extract the Local Optima Network (LON) of the QAP
 *
 * Source   : QAPnetwork.cpp
 * Language : C++
 * Author   : Sebastien Verel
 * Version  : 0.1
 * Date     : 07/30/2010
 *
 * Related publications:
 *
 * Daolio F., Tomassini M., Verel S., Ochoa G.
 * Communities of Minima in Local Optima Networks of Combinatorial Spaces
 * Physica A: Statistical Mechanics and its Applications, Volume 390, Issue 9, Pages 1684-1694, May 2011.
 *
 * Daolio F., Verel S., Ochoa G. & Tomassini M. (2010). 
 * Local Optima Networks of the Quadratic Assignement Problem. 
 * WCCI 2010 IEEE World Congress on Computational Intelligence (pp. 3145-3152). IEEE Press.
 *
 *
 * To compile:
 *   The GNU Scientific Library (GSL) is required. See the web site : http://www.gnu.org/s/gsl/ 
 * 
 *   Then type the classical commands:
 *      g++ -c qapNetwork.cpp -I.
 *      g++ -o qapNetwork qapNetwork.o -L. -lm -lgsl
 *
 * To execute:
 *   ./qapNetwork instances/KCso7uni-1.dat out.nodes out.edges   
 *   or
 *   ./qapNetwork instances/KCso7uni-1.dat out.nodes out.edges out.fitness
 *
 * "instances/KCso7uni-1.dat" is an instance file name of the QAP problem to consider in QAPlib format
 * some instances are given in the directory instances
 *
 * "out.nodes" is the file with the id, - fitness, and the basin size of each node (local optimum) of the network :
 *   idNode1 fitness1 basinSize1
 *   idNode2 fitness2 basinSize2
 *   ....
 * warning: the computed fitness is the opposite fitness (- fitness) to tranform the QAP into a maximization problem
 *
 * "out.edges" is the file with weighted edges of the network:
 *   idNode_i1 idNode_j1 weight_ij1 idNode_j2 wij2 idNode_j3 weight_ij3 ...
 *   ....
 *
 *  which means that there is the edge (idNode_i1, idNode_j1) with the weight weight_ij1, etc.
 *
 *
 * Warnings: 
 *   In this version, the full enumeration of the search space is computed.
 *   The search space have to be 'small'.
 *
 * This code could be used for other permutation search space (TSP, Flow shop, etc.)
 *
 * Contact:
 *   Sebastien Verel, sebastien.verel@unice.fr
 *
 **************************************************************************/

#ifndef __indexPerm
#define __indexPerm

/*
  class to compute of a solution for the QAP 
  
  The search space is the set of permutations
  This class makes 'linear' the search space
*/
class IndexPerm {
public:
    typedef unsigned short * Perm;

    /**
    constructor all permutation of size n
    
    @param n size of the permutation
    */
    IndexPerm(unsigned _n) : n(_n) {
	// number of permutations
	facto = new unsigned[n];

	size = fact(n);

	// creation du tableau
	permutations = new Perm[size];

	for(unsigned i = 0; i < size; i++)
	    permutations[i] = new unsigned short[n];

	// permutation initiale
	Perm t = new unsigned short[n];

	for(int j = 0; j < n; j++)
	    t[j] = j;

	create(0, t, n);
    }

    /**
    Destructor
    Time complexity O(n!)
    */
    ~IndexPerm() {
	for(unsigned i = 0; i < size; i++)
	    delete [] permutations[i];

	delete [] permutations;

	delete [] facto;
    }
    
    /**
    to have the permutation of a given index

    @param the index of the permutation

    @return the i th permutation
    */
    Perm getPerm(unsigned i) {
	return permutations[i] ;
    }


    /**
      to have the index of a given permutation

    @param p the permutation
    
    @return the index of the given permutation
    */
    unsigned getIndex(Perm p) {
	return recherche(p, 0, size - 1);
    }

    // total number of the permutations
    unsigned size;

    // size of the permutation
    unsigned n;

    // vector of all permutations in increasing order
    Perm * permutations;

private:
    // pre-computation of the factorial numbers
    unsigned * facto ;

    /**
    compute the factorial 

    @param k input number

    @return k!
    */
    unsigned fact(unsigned k) {
	facto[0] = 1;

	for(unsigned i = 1; i <= k; i++)
	    facto[i] = facto[i - 1] * i;
	
	return facto[k];
    }


    /* 
    Create the vector of all permutations

    @param k current index in the vector
    @param t current permutation
    @param p index in the permutation p
    */
    void create(unsigned k, Perm t, unsigned p) {
	if (p == 1) {
	    permutations[k][n - 1] = t[0];
	} else {
	    unsigned i, j, base, a;

	    unsigned fpm1 = facto[p - 1];

	    for(i = 0; i < p; i++) {
		base = k + i * fpm1;

		for(j = 0; j < fpm1; j++) 
		    permutations[base + j][n - p] = t[i]; 

		// copy the vector t without t[i] in increasing order
		Perm tt = new unsigned short[p - 1];
		a = 0;
		for(j = 0; j < p; j++)
		    if (i != j)
			tt[a++] = t[j];

		// recursive call
		create(base, tt, p - 1);

		delete [] tt ;
	    }
	}
    }

    /**
    Comparaison between permutation based on the lexicographic order
    if : 
       a < b -> +1
       a == b -> 0
       a > b -> -1

    @param p1 first permutation
    @param p2 second permutaiton

    @return comparaison value
    */
    int compare(Perm p1, Perm p2) {
	unsigned i = 0;

	while (i < n && p1[i] == p2[i])  i++;

	if (i == n)
	    return 0;
	else {
	    if (p1[i] < p2[i])
		return 1;
	    else
		return -1;
	}
    }

    /**
    Find the index of the permutation p in the vector of permutation

    @param p permutation to find
    @param a lower index
    @param b upper index

    @return index of the permutation p
    */
    unsigned recherche(Perm p, unsigned a, unsigned b) {
	unsigned c = (a + b) / 2;

	int r = compare(p, permutations[c]);

	if (r == 0)
	    return c;
	else
	    if (r == 1)
		return recherche(p, a, c - 1);
	    else
		return recherche(p, c + 1, b);
    }

};


#endif

