/**************************************************************************
 *
 * Source code to extract the Local Optima Network (LON) of the QAP
 *
 * Source   : QAPnetwork.cpp
 * Language : C++
 * Author   : Sebastien Verel
 * Version  : 0.1
 * Date     : 07/30/2010
 *
 * Related publications:
 *
 * Daolio F., Tomassini M., Verel S., Ochoa G.
 * Communities of Minima in Local Optima Networks of Combinatorial Spaces
 * Physica A: Statistical Mechanics and its Applications, Volume 390, Issue 9, Pages 1684-1694, May 2011.
 *
 * Daolio F., Verel S., Ochoa G. & Tomassini M. (2010). 
 * Local Optima Networks of the Quadratic Assignement Problem. 
 * WCCI 2010 IEEE World Congress on Computational Intelligence (pp. 3145-3152). IEEE Press.
 *
 *
 * To compile:
 *   The GNU Scientific Library (GSL) is required. See the web site : http://www.gnu.org/s/gsl/ 
 * 
 *   Then type the classical commands:
 *      g++ -c qapNetwork.cpp -I.
 *      g++ -o qapNetwork qapNetwork.o -L. -lm -lgsl
 *
 * To execute:
 *   ./qapNetwork instances/KCso7uni-1.dat out.nodes out.edges   
 *   or
 *   ./qapNetwork instances/KCso7uni-1.dat out.nodes out.edges out.fitness
 *
 * "instances/KCso7uni-1.dat" is an instance file name of the QAP problem to consider in QAPlib format
 * some instances are given in the directory instances
 *
 * "out.nodes" is the file with the id, - fitness, and the basin size of each node (local optimum) of the network :
 *   idNode1 fitness1 basinSize1
 *   idNode2 fitness2 basinSize2
 *   ....
 * warning: the computed fitness is the opposite fitness (- fitness) to tranform the QAP into a maximization problem
 *
 * "out.edges" is the file with weighted edges of the network:
 *   idNode_i1 idNode_j1 weight_ij1 idNode_j2 wij2 idNode_j3 weight_ij3 ...
 *   ....
 *
 *  which means that there is the edge (idNode_i1, idNode_j1) with the weight weight_ij1, etc.
 *
 *
 * Warnings: 
 *   In this version, the full enumeration of the search space is computed.
 *   The search space have to be 'small'.
 *
 * This code could be used for other permutation search space (TSP, Flow shop, etc.)
 *
 * Contact:
 *   Sebastien Verel, sebastien.verel@unice.fr
 *
 **************************************************************************/
#ifndef __pileH
#define __pileH


  /***************************************************************
   *
   * simple stack of unsigned int
   * with pointer
   *
   ***************************************************************/

class _StackNode 
{
public:
  unsigned v;
  _StackNode * next;

};

class Pile
{
public:

  /***************************************************************
   *
   * empty constructor
   *
   ***************************************************************/
  Pile() 
  {
    head = NULL;
    _size = 0;
  };

  /***************************************************************
   *
   * destructor
   *
   ***************************************************************/
  ~Pile() 
  {
    while (!empty()) 
      pop();
  };

  /***************************************************************
   *
   * true if the stack is empty
   *
   ***************************************************************/
  bool empty() {
    return (head == NULL);
  }
    
  /***************************************************************
   *
   * delete the head of the stack if it is possible
   *
   ***************************************************************/
  void pop() {
    if (!empty()) {
      _StackNode * n = head->next;
      delete head;
      head = n;
      _size--;
    } else
      cerr << "Pile.pop: error empty stack" << endl;
  }

  /***************************************************************
   *
   * return the value on the top of the stack
   *
   ***************************************************************/
  unsigned top() {
    if (!empty())       
      return head->v;
    else
      cerr << "Pile.top: error empty stack" << endl;
  }

  /***************************************************************
   *
   * add on the top a value
   *
   ***************************************************************/
  void push(unsigned value) {
    _StackNode * node = new _StackNode;

    node->v    = value;
    node->next = head;

    head = node;

    _size++;
  }

  /***************************************************************
   *
   * size of the stack
   *
   ***************************************************************/
  unsigned size() {
    return _size;
  }

private:
  _StackNode * head;

  unsigned _size;

};

#endif
