/**************************************************************************
 *
 * Source code to extract the Local Optima Network (LON) of the QAP
 *
 * Source   : QAPnetwork.cpp
 * Language : C++
 * Author   : Sebastien Verel
 * Version  : 0.1
 * Date     : 07/30/2010
 *
 * Related publications:
 *
 * Daolio F., Tomassini M., Verel S., Ochoa G.
 * Communities of Minima in Local Optima Networks of Combinatorial Spaces
 * Physica A: Statistical Mechanics and its Applications, Volume 390, Issue 9, Pages 1684-1694, May 2011.
 *
 * Daolio F., Verel S., Ochoa G. & Tomassini M. (2010). 
 * Local Optima Networks of the Quadratic Assignement Problem. 
 * WCCI 2010 IEEE World Congress on Computational Intelligence (pp. 3145-3152). IEEE Press.
 *
 *
 * To compile:
 *   The GNU Scientific Library (GSL) is required. See the web site : http://www.gnu.org/s/gsl/ 
 * 
 *   Then type the classical commands:
 *      g++ -c qapNetwork.cpp -I.
 *      g++ -o qapNetwork qapNetwork.o -L. -lm -lgsl -lgslcblas
 *
 * To execute:
 *   ./qapNetwork instances/KCso7uni-1.dat out.nodes out.edges   
 *   or
 *   ./qapNetwork instances/KCso7uni-1.dat out.nodes out.edges out.fitness
 *
 * "instances/KCso7uni-1.dat" is an instance file name of the QAP problem to consider in QAPlib format
 * some instances are given in the directory instances
 *
 * "out.nodes" is the file with the id, - fitness, and the basin size of each node (local optimum) of the network :
 *   idNode1 fitness1 basinSize1
 *   idNode2 fitness2 basinSize2
 *   ....
 * warning: the computed fitness is the opposite fitness (- fitness) to tranform the QAP into a maximization problem
 *
 * "out.edges" is the file with weighted edges of the network:
 *   idNode_i1 idNode_j1 weight_ij1 idNode_j2 wij2 idNode_j3 weight_ij3 ...
 *   ....
 *
 *  which means that there is the edge (idNode_i1, idNode_j1) with the weight weight_ij1, etc.
 *
 *
 * Warnings: 
 *   In this version, the full enumeration of the search space is computed.
 *   The search space have to be 'small'.
 *
 * This code could be used for other permutation search space (TSP, Flow shop, etc.)
 *
 * Contact:
 *   Sebastien Verel, sebastien.verel@unice.fr
 *
 **************************************************************************/


// standard includes

#include <stdexcept>  // runtime_error 
#include <iostream>   // cout
#include <sstream>    // ostrstream, istrstream
#include <fstream>

typedef unsigned short * Perm;

#include <evalQAP.h>
#include <indexPerm.h>
#include <solutions.h>

// permutation size
unsigned n;

// to print a permutation
void print(Perm & p) {
  for(unsigned j = 0; j < n; j++)
    std::cout << p[j] ;
}
      
/**
   arg[1] (required) : input file name of the QAP instance (QAPlib format)
   arg[2] (required) : output file name with nodes of the Local Optima Network
   arg[3] (required) : output file name with edges of the Local Optima Network

   arg[4] (not required) : output file name of the fitness value of each solutions
 */
int main(int argc, char **argv)
{
  if (argc < 4) {
    cerr << "At least 3 arguments required." << std::endl << std::endl;
    cerr << "./qapNetwork instances.dat out.nodes out.edges out.fitness" << std::endl;
    cerr << "     instance.dat (required)     : input  file name of the QAP instance (QAPlib format)"  << std::endl;
    cerr << "     out.nodes    (required)     : output file name of the nodes of the Local Optima Network" << std::endl;
    cerr << "     out.edges    (required)     : output file name of the edges of the Local Optima Network" << std::endl;
    cerr << "     out.fitness  (not required) : output file name of the fitness value of each solutions" << std::endl;
    exit(1);
  }

  // evaluation function of the QAP defined from the instance file
  EvalQAP eval(argv[1]);

  // size of the permutation
  n = eval.n;

  // the search space of permutations
  IndexPerm index(n);

  // the evaluation function can be change
  Solutions solutions(index, eval);

  // Compute the neutral networks of the fitness landscape
  solutions.computeNeutralNetworks();

  // Compute the basin of attraction
  solutions.computeBasins();

  // Export the file of all solutions:
  // numOfSolution fitness numOfNN sizeofNN following by the list of numOfBasin probToGo ...
  if (argc == 5)
    solutions.printSolutions(argv[4]);

  // compute the graph with weighted edges
  // out.node : numOfBasin fitness size
  // out.edge : numOfBasin following by the list of edges with the weight 
  solutions.computeGraph(argv[2], argv[3]);

  /* See:
       Verel S., Daolio F., Tomassini M., Ochoa G., "Local Optima Networks with Escape Edges", 
       International Conference on Artificial Evolution (EA-2011), Angers : France (2011)
     for more details on this Local Optima Network with escape edges */
  // solutions.computeEscapeGraph(argv[2], argv[3]);

  return 1;
}
