/**************************************************************************
 *
 * Source code to extract the Local Optima Network (LON) of the QAP
 *
 * Source   : QAPnetwork.cpp
 * Language : C++
 * Author   : Sebastien Verel
 * Version  : 0.1
 * Date     : 07/30/2010
 *
 * Related publications:
 *
 * Daolio F., Tomassini M., Verel S., Ochoa G.
 * Communities of Minima in Local Optima Networks of Combinatorial Spaces
 * Physica A: Statistical Mechanics and its Applications, Volume 390, Issue 9, Pages 1684-1694, May 2011.
 *
 * Daolio F., Verel S., Ochoa G. & Tomassini M. (2010). 
 * Local Optima Networks of the Quadratic Assignement Problem. 
 * WCCI 2010 IEEE World Congress on Computational Intelligence (pp. 3145-3152). IEEE Press.
 *
 *
 * To compile:
 *   The GNU Scientific Library (GSL) is required. See the web site : http://www.gnu.org/s/gsl/ 
 * 
 *   Then type the classical commands:
 *      g++ -c qapNetwork.cpp -I.
 *      g++ -o qapNetwork qapNetwork.o -L. -lm -lgsl
 *
 * To execute:
 *   ./qapNetwork instances/KCso7uni-1.dat out.nodes out.edges   
 *   or
 *   ./qapNetwork instances/KCso7uni-1.dat out.nodes out.edges out.fitness
 *
 * "instances/KCso7uni-1.dat" is an instance file name of the QAP problem to consider in QAPlib format
 * some instances are given in the directory instances
 *
 * "out.nodes" is the file with the id, - fitness, and the basin size of each node (local optimum) of the network :
 *   idNode1 fitness1 basinSize1
 *   idNode2 fitness2 basinSize2
 *   ....
 * warning: the computed fitness is the opposite fitness (- fitness) to tranform the QAP into a maximization problem
 *
 * "out.edges" is the file with weighted edges of the network:
 *   idNode_i1 idNode_j1 weight_ij1 idNode_j2 wij2 idNode_j3 weight_ij3 ...
 *   ....
 *
 *  which means that there is the edge (idNode_i1, idNode_j1) with the weight weight_ij1, etc.
 *
 *
 * Warnings: 
 *   In this version, the full enumeration of the search space is computed.
 *   The search space have to be 'small'.
 *
 * This code could be used for other permutation search space (TSP, Flow shop, etc.)
 *
 * Contact:
 *   Sebastien Verel, sebastien.verel@unice.fr
 *
 **************************************************************************/

#ifndef __solutions
#define __solutions

#include <solution.h>
#include <pile.h>
#include <basin.h>
#include <abrBasin.h>
#include <bestNeighbor.h>

#include <gsl/gsl_matrix.h>
#include <gsl/gsl_vector.h>
#include <gsl/gsl_linalg.h>
#include <math.h>

// value which means that the fitness value is not computed
#define DEFAULTVALUE 1 
// the minimum value of the fitness
#define MINVALUE -2147483647 

/*
  describe all the solutions of the landscapes
*/
class Solutions {
public:
    Solutions(IndexPerm & _index, EvalQAP & _eval) : index(_index), eval(_eval) {
	size = index.size;
	n    = index.n;

	sols = new Solution[size];

	for(unsigned i = 0; i < size; i++) {
	    sols[i].fitness = DEFAULTVALUE;
	    sols[i].nn      = NULL;
	    sols[i].basins  = NULL;
	}

	nbNN     = 0;
	nbBasins = 0; 
    }

    ~Solutions() {
	delete [] sols;
    }

    // tabular of all solutions
    Solution * sols;

    // number of solution in sols
    unsigned size;

    // size of the permutations
    unsigned n;

    // number of neutral networks
    unsigned nbNN;

    // number of basins of attraction
    unsigned nbBasins;

// ******************************

    /********************************************
    * compute all the neutral networks of the landscapes
    * and create the NN local optima basins
    *
    * In : none
    *
    * Out: 
    *   - all the fitness values are compute here (sols[i].fitness)
    *   - all the sols[i].nn of sols are computed
    *   - the number of neutral networks nbNN is computed too
    *   - the basins are created
    *********************************************/
    void computeNeutralNetworks() {
	for(unsigned i = 0; i < size; i++)
	    if (sols[i].nn == NULL) {
		// create the neutral network
		sols[i].nn = new NeutralNetwork;

		sols[i].nn->id      = nbNN;
		sols[i].nn->fitness = sols[i].fitness;
		sols[i].nn->size    = 0;
		sols[i].nn->nbDoors = 0;

		// increases the total number of NN
		nbNN++;
	    
		// compute the fitness with full evaluation if necessary
		if (sols[i].fitness == DEFAULTVALUE)
		    sols[i].fitness = eval(index.permutations[i]);
		
		computeNeutralNetwork(i);
	    }
    }

    /********************************************
    * compute the neutral network of the solution num
    *
    * In : none
    *
    * Out: 
    *   - the fitness values of all solutions in this NN are computed
    *   - sols[num].nn of sols is computed
    *********************************************/
    void computeNeutralNetwork(unsigned num) {
	// the neutral network to explore
	NeutralNetwork * nn = sols[num].nn;

	// create a stack to recursive exploration of the NN
	Pile stack;
	stack.push(num);

	unsigned numS;

	// the decimal number of the neighbor solution
	unsigned numV;
	Perm p = new unsigned short[n];
	unsigned k;

	// to describe the moves
	unsigned i, j;

	// fitness of the NN
	int fit = sols[num].fitness;

	// to test the door
	bool isDoor;

	while(!stack.empty()) {
	    numS = stack.top();
	    stack.pop();
	    
	    // give the same NN
	    sols[numS].nn = nn;

	    // increment the number of solution in the NN
	    nn->size++;

	    // copy the permutation to find the neighbors
	    for(k = 0; k < n; k++)
		p[k] = index.permutations[numS][k];

	    // to test if numS is a door of the NN
	    isDoor = false;

	    // to explore the neighborhood of numS
	    for(i = 0; i < n - 1; i++) 
		for(j = i+1; j < n; j++) {
		    swap(p, i, j);
		    numV = index.getIndex(p);

		    // incremental evaluation if necessary
		    if (sols[numV].fitness == DEFAULTVALUE)
			sols[numV].fitness = sols[numS].fitness + eval.compute_delta(index.permutations[numS], i, j);

		    // 
		    if (sols[numV].fitness > fit)
			isDoor = true;

		    if (sols[numV].fitness == fit) // if the solution has the same fitness
			if (sols[numV].nn == NULL) // if the solution is not yet tested
			    stack.push(numV);
		
		    swap(p, i, j);
		}

	    if (isDoor)
		nn->nbDoors++;
	}

	delete [] p;
    }

    /********************************************
    * compute all the basins of attraction of the landscapes
    *
    * In : none (the fitness values are supposed to be computed
    *
    * Out: 
    *   - all the sols[i].basins of sols are computed
    *   - the number of basins of attraction nbBasin is computed too
    *********************************************/
    void computeBasins() {
	for(unsigned i = 0; i < size; i++)
	    if (sols[i].basins == NULL) {
		// for the computation of basins which solution i is belonging
		computeBasins(i);
	    }
    }

    /********************************************
    * compute the basins of the solution num
    *
    * In : none
    *
    * Out: 
    *   - sols[num].basins of sols is computed
    *********************************************/
    void computeBasins(unsigned num) {
	// create the list of basins of attraction
	sols[num].basins = new CoupleList;

	// the basins to compute
	CoupleList * basins = sols[num].basins;

	if (sols[num].nn->nbDoors == 0) { // the solution belong to a NN local optima
	    // the id of basin is the id of the NN 
	    basins->cons(sols[num].nn->id, 1.0);
	} else {
	    unsigned bestList[n * (n - 1) / 2];
	    unsigned nbBest;
	    int bestFit;

	    // compute solution with higher or equal fitness
	    bestNeighborhood(num, bestFit, nbBest, bestList);

	    if (sols[num].fitness < bestFit) {
		// the basin is the linear combinaison of the basins of all best neighbors
		double alpha = 1.0 / (double) nbBest;

		unsigned numV;

		for(unsigned i = 0; i < nbBest; i++) {
		    numV = bestList[i];

		    if (sols[numV].basins == NULL)
			computeBasins(numV);

		    // add the basin of the neighbor solution to the current basin
		    add(*(sols[num].basins), *(sols[numV].basins), alpha);
		}
	    } else {
		// the solution num belongs to a "super" neutral network (with fitness = bestFit)
		computeBasinNeutral(num, bestFit, nbBest, bestList);

//		cerr << "neutrality too strong: yet this case is not done, sorry." << std::endl;
//		exit(1);
	    }
	}
    }

  /***************************************************************
  * compute all the best solution in the neighborhood of solution "num"
  * it uses the incremental evaluation 
  *
  * In:
  *    - num : the solution number
  * 
  * Out:
  *    - bestFit : the highest gitness value in the neighborhhod
  *    - nbBest  : number of neighbors which have bestFit fitness value
  *    - best    : list of the number of solutions (nbBest solutions)
  ***************************************************************/
    void bestNeighborhood(unsigned num, 
			  int & bestFit, unsigned & nbBest, unsigned best[]) {
	Perm p   = index.permutations[num];
	int  fit = sols[num].fitness;
	int  fitV;

	unsigned besti[n * (n - 1) / 2];
	unsigned bestj[n * (n - 1) / 2];

	unsigned i, j;

	nbBest = 0;

	// becarefull,  I suppose that the best fitness is higher than MINVALUE !
	bestFit = MINVALUE;

	// explore the neighborhood to find the best neighbors
	for(i = 0; i < n - 1; i++) 
	    for(j = i+1; j < n; j++) {
		fitV = fit + eval.compute_delta(p, i, j);

		// to compute best neighbors
		if (fitV > bestFit) {
		    bestFit  = fitV;
		    besti[0] = i;
		    bestj[0] = j;
		    nbBest   = 1;
		} else {
		    if (fitV == bestFit) {
			besti[nbBest] = i;
			bestj[nbBest] = j;
			nbBest++;
		    }
		}
	    }

	// find the numer of neighbors
	unsigned k;

	// copy the permutation to find the neighbors
	Perm pv = new unsigned short[n];
	for(k = 0; k < n; k++)
	    pv[k] = index.permutations[num][k];
	
	for(k = 0; k < nbBest; k++) {
	    swap(pv, besti[k], bestj[k]);
	    best[k] = index.getIndex(pv);
	    swap(pv, besti[k], bestj[k]);
	}
	delete [] pv;
	    
    }

  /***************************************************************
  * compute the number of solution with the same fitness in the neighborhood
  * it uses the incremental evaluation 
  *
  * In:
  *    - num : the solution number
  * 
  * Out:
  *    - nd : the neutral degree of the solution
  ***************************************************************/
    unsigned neutralDegree(unsigned num) {
	Perm p   = index.permutations[num];
	
	unsigned i, j;

	unsigned nd = 0;
	
	// explore the neighborhood 
	for(i = 0; i < n - 1; i++) 
	    for(j = i+1; j < n; j++) {
		if (eval.compute_delta(p, i, j) == 0)
		    nd++;
	    }

	return nd ;
    }

  /***************************************************************
   *
   * add the basin of attraction "basinV" (with coefficient alpha)
   * to basin "basin"
   * 
   *  keeping the order of "n" of the list basin
   *
   ***************************************************************/
  void add(CoupleList & basin, CoupleList & basinV, double alpha) {
    // place the iterator at the begining of the lists
    basin.initIterator();
    basinV.initIterator();

    // while there is something to add to basin
    while (!basinV.emptyIterator()) {
      if (basin.emptyIterator()) {       // basin is empty, so add node of basinV to basin 
	basin.consIterator(basinV.iterator()->n, alpha * basinV.iterator()->p);
	basinV.nextIterator();
      } else
	if (basinV.iterator()->n < basin.iterator()->n) { // the value of basinV is lower 
	  basin.consIterator(basinV.iterator()->n, alpha * basinV.iterator()->p);
	  basinV.nextIterator();
	} else
	  if (basin.iterator()->n == basinV.iterator()->n) {
	    basin.iterator()->p += alpha * basinV.iterator()->p;
	    basin.nextIterator();
	    basinV.nextIterator();
	  } else {
	    basin.nextIterator();
	  }
    }

  }

    
  /***************************************************************
  *
  * compute the basin of attractions of solution num
  * the solution num belongs to a "super" neutral network (with fitness = bestFit)
  *
  * In:
  *   - num: number of the solution
  *
  ***************************************************************/
    void computeBasinNeutral(unsigned num, int & bestFit, unsigned & nbBest, unsigned best[]) {

	// compute the "super" neutral network
	vector<BestNeighbor *> bestNetwork;
	Abr<BestNeighbor> inNetwork;
	Abr<BestNeighbor> outNetwork;
      
	computeSuperNetwork(num, bestNetwork, inNetwork, outNetwork);

	// compute the nodes out (in the neighborhood) of the "super" neutral network
	vector<BestNeighbor *> bestOut;
	outNetwork.get(bestOut);
      
	for(unsigned i = 0; i < bestOut.size(); i++)
	    if (sols[bestOut[i]->id].basins == NULL)
		computeBasins(bestOut[i]->id);

	if (inNetwork.size() == 1) {
	    double alpha = 1.0 / (double) nbBest;
	    
	    unsigned numV;
	    
	    for(unsigned i = 0; i < nbBest; i++) {
		numV = best[i];

		if (sols[numV].basins == NULL)
		    computeBasins(numV);

		add(*(sols[num].basins), *(sols[numV].basins), alpha);
	    }	
	} else {
	    vector<BestNeighbor *> networkList;
	    inNetwork.get(networkList);

	    // the list of attractors (number of neutral network)
	    Abr<BestNeighbor> attractors;
	    vector<BestNeighbor *> attractorsList;
	    computeAttractors(bestOut, attractors, attractorsList);
      
	    gsl_matrix * A =  gsl_matrix_calloc(inNetwork.size(), inNetwork.size());

	    vector<gsl_vector *> outLinks(attractors.size());
	    for(unsigned i = 0; i < attractors.size(); i++)
		outLinks[i] = gsl_vector_calloc(inNetwork.size());

	    computeMatriceBestNetwork(networkList, inNetwork, outNetwork, attractors, 
				      A, outLinks);

	    // find the probability for each basin of attraction
	    gsl_vector * pBasin = gsl_vector_alloc(inNetwork.size());

	    int s;
	
	    gsl_permutation * p = gsl_permutation_alloc (inNetwork.size());     
      
	    unsigned j;
	    gsl_linalg_LU_decomp(A, p, &s);
     
	    for(int i = attractors.size() - 1; i >= 0 ; i--) {
		// compute the probability to acces to basin of attraction i]

		// solve the linear equation
		gsl_linalg_LU_solve(A, p, outLinks[i], pBasin);     

		// add this probability to each node of the super neutral network
		for(j = 0; j < networkList.size(); j++) {
		    if (sols[ networkList[j]->id ].basins == NULL)
			sols[ networkList[j]->id ].basins = new CoupleList;

		    sols[ networkList[j]->id ].basins->cons(attractorsList[i]->id, gsl_vector_get(pBasin, j)); 
		}
	    }
      
	    gsl_matrix_free(A);
      
	    for(unsigned i = 0; i < attractors.size(); i++)
		gsl_vector_free(outLinks[i]);
	}
    }

  /***************************************************************
   *
   * compute the "super" neutral network and its neighborhood
   *
   * 
   *
   ***************************************************************/
  void computeSuperNetwork(unsigned num, 
			   vector<BestNeighbor *> & bestNetwork,
			   Abr<BestNeighbor> & inNetwork,
			   Abr<BestNeighbor> & outNetwork) {
      Pile stack;
      stack.push(num);

      unsigned best[n * (n - 1) / 2];
      unsigned nbBest;
      int bestFit;

      CoupleList * bestList;

      BestNeighbor * node;

      // the decimal number of the neighbor solution
      unsigned i;

      while(!stack.empty()) {
	  num = stack.top();
	  stack.pop();
	  
	  // compute solution whith higher or equal fitness
	  bestNeighborhood(num, bestFit, nbBest, best);

	  bestList = new CoupleList();
      
	  for(i = 0; i < nbBest; i++)
	      bestList->consOrder(best[i], 1.0 / nbBest);  // add in increase order of number complexity O(n)
      
	  node = new BestNeighbor;
	  node->set(num, bestFit, bestList);

	  bestNetwork.push_back(node);
	  
	  if (bestFit == sols[num].fitness) { // if the best solution has the same fitness
	      inNetwork.add(node);
	      
	      for(i = 0; i < nbBest; i++)  // then the best neighbors belong to the "super" netral network
		  if (!isInSuperNetwork(best[i], inNetwork, outNetwork)) { // complexity O(log(n))
		      stack.push(best[i]);
		  }
	  } else
	      outNetwork.add(node);
	  
      }
  }
    
  /***************************************************************
   *
   * true if numV is in the super network
   *
   *
   ***************************************************************/
  bool isInSuperNetwork(unsigned numV, 
			Abr<BestNeighbor> & inNetwork,
			Abr<BestNeighbor> & outNetwork) {
    if (inNetwork.find(numV))
      return true;
    else
      return (outNetwork.find(numV));
  }

  /***************************************************************
   *
   * compute the set of attractors obtain by the super network
   * (remind that a attractor is a number of neutral network)
   *
   ***************************************************************/
  void computeAttractors(vector<BestNeighbor *> & bestOut, 
			 Abr<BestNeighbor> & attractors,
			 vector<BestNeighbor *> & attractorsList) {
    
    unsigned num, nr;

    BestNeighbor * node;

    for(unsigned i = 0; i < bestOut.size(); i++) {
      num = bestOut[i]->id;

      sols[num].basins->initIterator();
      while (!sols[num].basins->emptyIterator()) {
	nr  = sols[num].basins->iterator()->n;
	
	node = new BestNeighbor;
	node->set(nr, 0.0, NULL);
	attractors.add(node);

	sols[num].basins->nextIterator();
      }
    }

    attractors.get(attractorsList);
  }

  /***************************************************************
   *
   * compute the adjacence matrix of best neutral network and the list of solution
   *
   *
   ***************************************************************/
  void computeMatriceBestNetwork(vector<BestNeighbor *> & networkList, 
				 Abr<BestNeighbor> & inNetwork,
				 Abr<BestNeighbor> & outNetwork,
				 Abr<BestNeighbor> & attractors,
				 gsl_matrix * A, 
				 vector<gsl_vector *> & outLinks
				 ) {
    unsigned i, j;
    unsigned k, l;

    BestNeighbor * node;
    BestNeighbor * pAttractor;
    
    double prob, ancProb, alpha;

    unsigned nr, numV;

    BestNeighbor * p;

    for(i = 0; i < networkList.size(); i++) {
      // diagonal
      gsl_matrix_set(A, i, i, 1.0);

      node = networkList[i];
      alpha = 1.0 / node->best->size();

      node->best->initIterator();
      while(!node->best->emptyIterator()) {
	numV = node->best->iterator()->n;

	p = inNetwork.find(numV);

	if (p != NULL) { // numV is in the super network
	  j = p->position;
	  gsl_matrix_set(A, i, j, - node->best->iterator()->p);
	} else {
	    // parcourir le vecteur de prob du basin

	    sols[numV].basins->initIterator();
	    while (!sols[numV].basins->emptyIterator()) {
		nr   = sols[numV].basins->iterator()->n;
		prob = sols[numV].basins->iterator()->p;

		pAttractor = attractors.find(nr);
		ancProb = gsl_vector_get(outLinks[pAttractor->position], i);
		gsl_vector_set(outLinks[pAttractor->position], i, ancProb + alpha * prob);
		
		sols[numV].basins->nextIterator();
	    }

	}
	
	node->best->nextIterator();
      }
    } 

  }

// ***********************************
    /* compute the graph itself : nodes and weighted edges */

    /******************************
    * to compute the network with weighted edges:
    *    basin: best improvement
    *    edges: based on random solution of basin
    *
    * In:
    *   - files name of nodes and edges
    * Out:
    *   - files are filed
    ******************************/
    void computeGraph(char * fileNameNodes, char * fileNameEdges) {
	fstream fileNode(fileNameNodes, ios::out);
	fstream fileEdge(fileNameEdges, ios::out);
	
	if (fileNode.is_open() && fileEdge.is_open()) {
	    // parcours tous les solutions pour faire le bilan des basins existant
	    Abr<Basin> abr ;

	    unsigned s, i, j, vi, vj, k;

	    Basin * bi;

	    // number of neighbor of solution s
	    unsigned nj ;

	    CoupleList ** bj = new CoupleList*[n * (n - 1) / 2];

	    // to copy the solution
	    Perm pv = new unsigned short[n];

	    // for all the solution
	    for(unsigned s = 0; s < size; s++) {

		// save all the basin pointer of neighbor's solution of s
		for(k = 0; k < n; k++)
		    pv[k] = index.permutations[s][k];

		nj = 0;		
		for(vi = 0; vi < n - 1; vi++)
		    for(vj = vi + 1; vj < n; vj++) {
			swap(pv, vi, vj);
			bj[nj] = sols[index.getIndex(pv)].basins;
			swap(pv, vi, vj);
			nj++;
		    }

		// increase of basin of solution k
		CoupleList * basins = sols[s].basins;
		basins->initIterator();
		// while there is a basins for solution s
		while (!basins->emptyIterator()) {
		    bi = abr.find(basins->iterator()->n);
    
		    if (bi == NULL) {  // this basin does not exist in the tree
			bi          = new Basin;
			bi->id      = basins->iterator()->n;
			bi->fitness = DEFAULTVALUE;
			bi->size    = 0;

			abr.add(bi);
		    }

		    // to set the fitness of the basin
		    if (basins->iterator()->n == sols[s].nn->id)
			bi->fitness = sols[s].nn->fitness;

		    // increase the size of bi
		    bi->size += basins->iterator()->p;

		    // for all the neighbor s^' of s compute \sum pi(s) * pj(s^')
		    for(j = 0 ; j < nj; j++) 
			add(bi->edges, *(bj[j]), basins->iterator()->p);

		    basins->nextIterator();
		}
		
	    } // end for all solution s	    

	    // put all the informations into files
	    vector<Basin *> theBasins;
	    abr.get(theBasins);

	    // initialise the number of basins
	    nbBasins = theBasins.size();

	    for(i = 0; i < nbBasins; i++) {
		// informations on the nodes
		fileNode << theBasins[i]->id      << " ";
		fileNode << theBasins[i]->fitness << " ";
		fileNode << theBasins[i]->size    << std::endl;

		// information on the edges
		fileEdge << theBasins[i]->id ;

		theBasins[i]->edges.initIterator();
		while (!theBasins[i]->edges.emptyIterator()) {
		    fileEdge << " " <<  theBasins[i]->edges.iterator()->n ;
		    fileEdge << " " << (theBasins[i]->edges.iterator()->p / (theBasins[i]->size * (n * (n - 1) / 2))) ;

		    theBasins[i]->edges.nextIterator();
		}

		fileEdge << std::endl;
	    }

	    fileNode.close();
	    fileEdge.close();

	    // delete all the data
	    delete [] bj;
	    delete [] pv;

	    abr.del();
	} else
	    cerr << "Solutions::computeGraph: impossible to open " << fileNameNodes << " or " << fileNameEdges << std::endl;
    }

    /******************************
    * to compute the network with escape edges:
    *    basin: best improvement
    *    edges: from local optima and distance 1 (2-opt, swapping two locations)
    *
    * In:
    *   - files name of nodes and edges
    * Out:
    *   - files are filed
    ******************************/
    void computeEscapeGraph(char * fileNameNodes, char * fileNameEdges) {
	fstream fileNode(fileNameNodes, ios::out);
	fstream fileEdge(fileNameEdges, ios::out);
	
	if (fileNode.is_open() && fileEdge.is_open()) {
	    // parcours tous les solutions pour faire le bilan des basins existant
	    Abr<Basin> abr ;

	    unsigned s, i, j, vi, vj, k;

	    Basin * bi;

	    // number of neighbor of solution s
	    unsigned nj ;

	    CoupleList ** bj = new CoupleList*[n * (n - 1) / 2];

	    // to duplicate the solution in order to compute the neighbors
	    Perm pv = new unsigned short[n];

	    // for all the solutions
	    for(unsigned s = 0; s < size; s++) {

		// considered only the solutions which are in a local optima neutral network
		if (sols[s].nn->nbDoors == 0) {

		    // to compute the neighbor's of the solution, copy the current solution
		    for(k = 0; k < n; k++)
			pv[k] = index.permutations[s][k];

		    nj = 0;		
		    for(vi = 0; vi < n - 1; vi++)
			for(vj = vi + 1; vj < n; vj++) {
			    swap(pv, vi, vj);
			    bj[nj] = sols[index.getIndex(pv)].basins;
			    swap(pv, vi, vj);
			    nj++;
			}
		    
		    // increase of basin of solution k
		    CoupleList * basins = sols[s].basins;   // logiquement il n'y qu'un seul basin puisque la solution est dans unoptimum local
		    basins->initIterator();
		    // while there is a basins for solution s
		    while (!basins->emptyIterator()) {
			bi = abr.find(basins->iterator()->n);
			
			if (bi == NULL) {  // this basin does not exist in the tree
			    bi          = new Basin;
			    bi->id      = basins->iterator()->n;
			    bi->fitness = DEFAULTVALUE;
			    bi->size    = 0;
			    
			    abr.add(bi);
			}
			
			// to set the fitness of the basin
			if (basins->iterator()->n == sols[s].nn->id)
			    bi->fitness = sols[s].nn->fitness;
			
			// increase the size of bi
			bi->size += basins->iterator()->p;
			
			// for all the neighbor s^' of s compute \sum pi(s) * pj(s^')
			for(j = 0 ; j < nj; j++) 
			    add(bi->edges, *(bj[j]), basins->iterator()->p);
			
			basins->nextIterator();
		    }
		} else {
		    CoupleList * basins = sols[s].basins;   
		    basins->initIterator();
		    // while there is a basins for solution s
		    while (!basins->emptyIterator()) {
			bi = abr.find(basins->iterator()->n);
			
			if (bi == NULL) {  // this basin does not exist in the tree
			    bi          = new Basin;
			    bi->id      = basins->iterator()->n;
			    bi->fitness = DEFAULTVALUE;
			    bi->size    = 0;
			    
			    abr.add(bi);
			}
			
			// increase the size of bi
			bi->size += basins->iterator()->p;

			basins->nextIterator();
		    }
		}
	    } // end for all solution s	    

	    // put all the informations into files
	    vector<Basin *> theBasins;
	    abr.get(theBasins);

	    // initialise the number of basins
	    nbBasins = theBasins.size();

	    for(i = 0; i < nbBasins; i++) {
		// informations on the nodes
		fileNode << theBasins[i]->id      << " ";
		fileNode << theBasins[i]->fitness << " ";
		fileNode << theBasins[i]->size    << std::endl;

		// information on the edges
		fileEdge << theBasins[i]->id ;

		theBasins[i]->edges.initIterator();
		while (!theBasins[i]->edges.emptyIterator()) {
		    fileEdge << " " <<  theBasins[i]->edges.iterator()->n ;
		    fileEdge << " " << (theBasins[i]->edges.iterator()->p / (theBasins[i]->size * (n * (n - 1) / 2))) ;

		    theBasins[i]->edges.nextIterator();
		}

		fileEdge << std::endl;
	    }

	    fileNode.close();
	    fileEdge.close();

	    // delete all the data
	    delete [] bj;
	    delete [] pv;

	    abr.del();
	} else
	    cerr << "Solutions::computeEscapeGraph: impossible to open " << fileNameNodes << " or " << fileNameEdges << std::endl;
    }

// ***********************************
    /* output methods */

    /********************************************
    * compute all neutral degree of solution 
    *
    * In:
    *   - file name
    *
    * Out:
    *   - the file is filled
    *********************************************/
    void computeNeutralDegree(const char * fileName) {
	fstream file(fileName, ios::out);

	unsigned m = size;
	unsigned M = 0;
	double avg = 0.0;
	double sd  = 0.0;

	unsigned nd ;

	if (file.is_open()) {
	    for(unsigned i = 0; i < size; i++) {
//		file << neutralDegree(i) << std::endl;
		nd = neutralDegree(i) ;

		if (M < nd) M = nd;
		if (nd < m) m = nd;
		
		avg += nd;
		sd  += nd * nd;
	    }
	    

	    avg /= size;
	    
	    sd  = sd / size - avg * avg ;
	    if (sd > 0) 
		sd = sqrt(sd);
	    else
		sd = 0.0;
	    
	    file << avg << " " << sd << " " << m << " " << M << endl;
	}
	
	file.close();
    }

    /******************************
    * to print the solution into a file
    *
    * In:
    *   - file name
    *
    * Out:
    *   - the file is filled
    ******************************/
    void printSolutions(const char * fileName) {
	fstream file(fileName, ios::out);

	if (file.is_open()) {

	    for(unsigned i = 0; i < size; i++) {
		file << i << " " ;

                // fitness
		file << sols[i].fitness << " ";
		
		// neutral networks informations
		file << sols[i].nn->id      << " ";
		file << sols[i].nn->size    << " ";
		file << sols[i].nn->nbDoors << " ";

		// basins which belongs i
		sols[i].basins->print(file);

		file << std::endl;
	    }

	} else
	    cerr << "impossible to open " << fileName << std::endl;
    }

// ******************************
private:
    IndexPerm & index;

    EvalQAP & eval;

    void swap(Perm p, unsigned i, unsigned j) {
	unsigned short tmp = p[i];

	p[i] = p[j];
	p[j] = tmp;
    }
};

#endif

// Local Variables:
// coding: iso-8859-1
// mode: C++
// c-file-offsets: ((c . 0))
// c-file-style: "Stroustrup"
// fill-column: 80
// End:

